#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

###############################################################################
#                                                                             #
# Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.web module of the BEAT platform.              #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################

from django.conf import settings
from django.contrib.auth.models import User
from django.shortcuts import get_object_or_404
from django.http import Http404
from django.db import models
from django.db.models import Q
from django.core.urlresolvers import reverse

from rest_framework import generics
from rest_framework import views
from rest_framework import permissions
from rest_framework.response import Response
from rest_framework import status

from .serializers import FullSupervisionTrackSerializer
from .serializers import SupervisionTrackUpdateSerializer
from .serializers import UserSerializer

from ..common.utils import validate_restructuredtext
from ..ui.templatetags.markup import restructuredtext


from .models import SupervisionTrack, Profile
from ..common.models import Shareable
from ..common.exceptions import ShareError
from ..common.mixins import CommonContextMixin
from ..ui.registration.models import RegistrationProfile
from ..utils import mail

from itertools import chain
from datetime import datetime, timedelta


from .permissions import IsSupervisorAndAuthor, IsAuthorAndNotSupervisor, IsAuthorAndPossiblySupervisor, IsSuperUser

from ..common.responses import BadRequestResponse, ForbiddenResponse

import datetime
import re
try:
  from urlparse import urlparse
except ImportError:
  from urllib.parse import urlparse

import simplejson as json


#----------------------------------------------------------


class SupervisorListView(generics.ListAPIView):
    model = SupervisionTrack
    serializer_class = FullSupervisionTrackSerializer

    def get_permissions(self):
        permission_classes = [permissions.IsAuthenticated, IsSupervisorAndAuthor]

        self.permission_classes = permission_classes

        return super(SupervisorListView, self).get_permissions()


    def get_serializer(self, *args, **kwargs):

        return super(SupervisorListView, self).get_serializer(*args, **kwargs)


    def list(self, request):
        #A supervisor can validate an account of:
        #1) a new user requesting validation
        #2) an existing validated user rejected by a previous supervisor
        #3) an existing validated user requesting a change of supervisor
        #4) a blocked user requesting a supervision
        #On all cases check the current key in supervisee profile match the supervisiontrack key as this is the current supervision request/track from the supervisee
        queryset    = SupervisionTrack.objects.filter(supervisor=request.user).filter(Q(supervisee__profile__status=Profile.WAITINGVALIDATION)|Q(supervisee__profile__status=Profile.REJECTED)|Q(supervisee__profile__status=Profile.YEARREVALIDATION)|Q(supervisee__profile__status=Profile.ACCEPTED)|Q(supervisee__profile__status=Profile.BLOCKED)).filter(Q(supervisee__profile__supervision_key=models.F('supervision_key')))
        serializer  = FullSupervisionTrackSerializer(queryset, many=True, context ={'request': request})

        return Response(serializer.data)


#----------------------------------------------------------


class BaseUpdateSupervisionTrackView(generics.UpdateAPIView):
    model = SupervisionTrack
    serializer_class = SupervisionTrackUpdateSerializer

    def get_permissions(self):
        permission_classes = [permissions.IsAuthenticated, IsSupervisorAndAuthor]

        self.permission_classes = permission_classes

        return super(BaseUpdateSupervisionTrackView, self).get_permissions()


#----------------------------------------------------------


class SupervisorAddSuperviseeView(BaseUpdateSupervisionTrackView):
    permission_classes = BaseUpdateSupervisionTrackView.permission_classes

    def put(self, request, supervisee_name):
        supervisee = User.objects.get(username=supervisee_name)
        profile_supervisee = Profile.objects.get(user=supervisee)
        if supervisee.profile.status != Profile.ACCEPTED:
            if not request.user.is_superuser:
                supervisiontrack = SupervisionTrack.objects.get(supervisor=request.user, supervisee=supervisee, supervisee__profile__supervision_key=models.F('supervision_key'))
            else:
                try:
                    # admin that validates a supervisee
                    supervisiontrack = SupervisionTrack.objects.get(supervisor=request.user, supervisee=supervisee, supervisee__profile__supervision_key=models.F('supervision_key'))
                except:
                    # supervisor request for admin
                    supervisiontrack = SupervisionTrack.objects.create(
                        supervisee = supervisee,
                        supervisor = request.user,
                        is_valid = True,
                        )
                    #Assign key to supervision track
                    supervisiontrack.supervision_key = supervisee.profile.supervision_key

            supervisiontrack.is_valid = True

            now = datetime.datetime.now()
            expiration_date_delta = datetime.timedelta(days=settings.ACCOUNT_EXPIRATION_DAYS)
            supervisiontrack.expiration_date = now + expiration_date_delta

            supervisiontrack.start_date = now
            supervisiontrack.last_validation_date = now
            supervisee.profile.status = Profile.ACCEPTED
            supervisee.profile.rejection_date = None

            supervisiontrack.save()
            supervisee.profile.save()
            supervisee.is_active = True
            supervisee.save()

            parsed_url = urlparse(settings.URL_PREFIX)
            server_address = '%s://%s' % (parsed_url.scheme, parsed_url.hostname)

            context = {
                'supervisor': supervisiontrack.supervisor,
                'supervisee': supervisee,
                'prefix': server_address,
            }

            mail.send_email('registration/mail.supervisor_validated.subject.txt',
                            'registration/mail.supervisor_validated.message.txt',
                            context,
                            [supervisee.email])

            return Response(status=status.HTTP_204_NO_CONTENT)
        else:
            #Track already valid
            reason = "You can't validate this supervisee as his supervision is still valid"
            result = {
                'error': reason,
            }
            return BadRequestResponse(result)


#----------------------------------------------------------


class SupervisorRemoveSuperviseeView(BaseUpdateSupervisionTrackView):
    permission_classes = BaseUpdateSupervisionTrackView.permission_classes

    def put(self, request, supervisee_name):
        supervisee = User.objects.get(username=supervisee_name)
        if not supervisee.profile.is_supervisor:
            supervisiontrack = SupervisionTrack.objects.get(supervisor=request.user, supervisee=supervisee, supervisee__profile__supervision_key=models.F('supervision_key'))
        else:
            if not supervisee.profile.status == Profile.WAITINGVALIDATION:
                try:
                    supervisiontrack = SupervisionTrack.objects.get(supervisor=request.user, supervisee=supervisee, supervisee__profile__supervision_key=models.F('supervision_key'))
                except:
                    if supervisee.profile.is_supervisor and request.user.is_superuser:
                        supervisiontrack = SupervisionTrack.objects.get(supervisee=supervisee, supervisee__profile__supervision_key=models.F('supervision_key'))
                    else:
                        reason = "You can't remove this supervisor"
                        result = {
                            'error': reason,
                        }
                        return BadRequestResponse(result)


        parsed_url = urlparse(settings.URL_PREFIX)
        server_address = '%s://%s' % (parsed_url.scheme, parsed_url.hostname)

        if not supervisee.profile.is_supervisor:
            context = {
                'supervisor': supervisiontrack.supervisor,
                'supervisee': supervisee,
                'prefix': server_address,
            }
        else:
            context = {
                'supervisor': request.user,
                'supervisee': supervisee,
                'prefix': server_address,
            }

        if supervisee.profile.status == Profile.WAITINGVALIDATION:
            #New user account waiting validation, so delete this account and inform by email the user
            mail.send_email('registration/mail.supervisor_rejected.subject.txt',
                            'registration/mail.supervisor_rejected_delete_account.message.txt',
                            context,
                            [supervisee.email])

            registration_profile = RegistrationProfile.objects.get(user=supervisee)
            supervisee.profile.delete()
            supervisee.delete()
            if not supervisee.profile.is_supervisor:
                supervisiontrack.delete()
            registration_profile.delete()
        else:
            #Reject this account and inform by email the user
            now = datetime.datetime.now()
            expiration_date_delta = datetime.timedelta(days=settings.ACCOUNT_BLOCKAGE_AFTER_FIRST_REJECTION_DAYS)

            supervisiontrack.expiration_date = now
            supervisiontrack.is_valid = False

            if supervisee.profile.status != Profile.BLOCKED:
                supervisee.profile.status = Profile.REJECTED
                if supervisee.profile.rejection_date == None:
                    supervisee.profile.rejection_date = now + expiration_date_delta
            else:
                    supervisee.profile.rejection_date = None

            supervisee.profile.supervision_key = None

            supervisiontrack.save()
            supervisee.profile.save()
            supervisee.save()

            if not supervisee.profile.is_supervisor:

                mail.send_email('registration/mail.supervisor_rejected.subject.txt',
                                'registration/mail.supervisor_rejected.message.txt',
                                context,
                                [supervisee.email])
            else:

                now = datetime.datetime.now()
                expiration_date_delta = datetime.timedelta(days=settings.ACCOUNT_BLOCKAGE_AFTER_FIRST_REJECTION_DAYS)

                # Transform supervisor account to normal account and inform by email
                supervisee.profile.is_supervisor = False
                supervisee.profile.save()
                supervisee.save()

                mail.send_email('registration/mail.beatadmin_supervisor_rejected.subject.txt',
                                'registration/mail.beatadmin_supervisor_rejected.message.txt',
                                context,
                                [supervisee.email])

                # Reject all his supervisees and inform by email
                # Make sure all supervision tracks are invalid for rejected supervisor
                supervisiontracks = SupervisionTrack.objects.filter(supervisor=supervisee).filter(Q(supervisee__profile__status=Profile.REJECTED)|Q(supervisee__profile__status=Profile.YEARREVALIDATION)|Q(supervisee__profile__status=Profile.ACCEPTED)|Q(supervisee__profile__status=Profile.BLOCKED)).filter(Q(supervisee__profile__supervision_key=models.F('supervision_key')))
                supervisiontracks_new_users = SupervisionTrack.objects.filter(supervisor=supervisee).filter(Q(supervisee__profile__status=Profile.NEWUSER)|Q(supervisee__profile__status=Profile.WAITINGVALIDATION))

                if supervisiontracks.count() > 0:
                    for track in supervisiontracks:
                        track.expiration_date = now
                        track.is_valid = False

                        track_supervisee = track.supervisee
                        if track_supervisee.profile.status != Profile.BLOCKED:
                            track_supervisee.profile.status = Profile.REJECTED
                            if track_supervisee.profile.rejection_date == None:
                                track_supervisee.profile.rejection_date = now + expiration_date_delta
                        else:
                                track_supervisee.profile.rejection_date = None

                        track_supervisee.profile.supervision_key = None

                        track.save()
                        track_supervisee.profile.save()
                        track_supervisee.save()

                        context = {
                            'supervisor': supervisee,
                            'supervisee': track_supervisee,
                            'prefix': server_address,
                        }

                        mail.send_email('registration/mail.supervisor_failed_rejected.subject.txt',
                                        'registration/mail.supervisor_failed_rejected.message.txt',
                                        context,
                                        [track_supervisee.email])

                        if track_supervisee.profile.status == Profile.WAITINGVALIDATION:
                            registration_profile = RegistrationProfile.objects.get(user=track_supervisee)
                            track_supervisee.profile.delete()
                            track_supervisee.delete()
                            registration_profile.delete()

                        #New supervision request refused
                        if track.start_date is None:
                            track.delete()

                if supervisiontracks_new_users.count() > 0:
                    for track in supervisiontracks_new_users:
                        track_supervisee = track.supervisee
                        track_supervisor = track.supervisor

                        context = {
                            'supervisor': track_supervisor,
                            'supervisee': track_supervisee,
                            'prefix': server_address,
                        }

                        #New user account waiting validation, so delete this account and inform by email the user
                        mail.send_email('registration/mail.supervisor_rejected.subject.txt',
                                        'registration/mail.supervisor_rejected_delete_account.message.txt',
                                        context,
                                        [track_supervisee.email])

                        registration_profile = RegistrationProfile.objects.get(user=track_supervisee)
                        track_supervisee.profile.delete()
                        track_supervisee.delete()
                        registration_profile.delete()

                        #New supervision request refused
                        if track.start_date is None:
                            track.delete()

            #New supervision request refused
            if supervisiontrack.start_date is None:
                supervisiontrack.delete()

        return Response(status=status.HTTP_204_NO_CONTENT)


#----------------------------------------------------------


class BaseCreateSupervisionTrackViewSupervisee(generics.CreateAPIView):
    model = SupervisionTrack
    serializer_class = SupervisionTrackUpdateSerializer

    def get_permissions(self):
        permission_classes = [permissions.IsAuthenticated, IsAuthorAndNotSupervisor]

        self.permission_classes = permission_classes

        return super(BaseCreateSupervisionTrackViewSupervisee, self).get_permissions()


#----------------------------------------------------------


class SuperviseeAddSupervisorView(BaseCreateSupervisionTrackViewSupervisee):
    permission_classes = BaseCreateSupervisionTrackViewSupervisee.permission_classes

    def post(self, request, supervisor_name):
        #check if user exists and its validity and if it's a supervisor account
        try:
            supervisor = User.objects.get(username=supervisor_name)
            if not supervisor.profile.is_supervisor:
                #Not a valid supervisor
                reason = "Not a valid supervisor request"
                result = {
                    'error': reason,
                }
                return BadRequestResponse(result)
            else:
                if supervisor.profile.status == Profile.BLOCKED:
                    #Not a valid supervisor
                    reason = "Not a valid supervisor request"
                    result = {
                        'error': reason,
                    }
                    return BadRequestResponse(result)
        except:
            #This username does not exist but don't give too much information
            reason = "Not a valid supervisor request"
            result = {
                'error': reason,
            }
            return BadRequestResponse(result)

        parsed_url = urlparse(settings.URL_PREFIX)
        server_address = '%s://%s' % (parsed_url.scheme, parsed_url.hostname)

        supervisee = request.user
        if supervisee.profile.supervision_key is not None:
            #There's a key check if there's a valid track
            supervisiontrack = SupervisionTrack.objects.get(supervisee=supervisee, supervisee__profile__supervision_key=models.F('supervision_key'))
            if supervisiontrack.is_valid:
                if supervisee.profile.status != Profile.WAITINGVALIDATION and supervisee.profile.status != Profile.NEWUSER and supervisee.profile.status != Profile.BLOCKED:
                    #Stop the current supervision
                    now = datetime.datetime.now()
                    expiration_date_delta = datetime.timedelta(days=settings.ACCOUNT_BLOCKAGE_AFTER_FIRST_REJECTION_DAYS)


                    supervisiontrack.expiration_date = now
                    supervisiontrack.is_valid = False
                    supervisiontrack.save()

                    #Inform by email the revoked supervisor
                    context = {
                        'supervisor': supervisiontrack.supervisor,
                        'supervisee': supervisiontrack.supervisee,
                        'prefix': server_address,
                    }

                    mail.send_email('registration/mail.supervisor_rejection.subject.txt',
                                    'registration/mail.supervisor_rejection.message.txt',
                                    context,
                                    [supervisiontrack.supervisor.email])

                    #Reject this account and inform by email the supervisor and the user

                    supervisee.profile.status = Profile.REJECTED
                    supervisee.profile.supervision_key = None
                    if supervisee.profile.rejection_date == None:
                        supervisee.profile.rejection_date = now + expiration_date_delta

                    supervisee.profile.save()
                    supervisee.save()
                else:
                    #Not allowed to do this (unproper profile.status)
                    reason = "You are not able to perform this action, your profile is %s"%(supervisee.profile.status)
                    result = {
                        'error': reason,
                    }
                    return BadRequestResponse(result)

            else:
                #A pending request already exist
                reason = "You are not able to perform this action as you already have a pending request"
                result = {
                    'error': reason,
                }
                return BadRequestResponse(result)
        else:
            #No key is present in supervisee

            #Make sure all tracks are invalid
            supervisiontracks = SupervisionTrack.objects.filter(supervisee=supervisee, is_valid=True)
            # This should never be the case but if it happens invalidate all tracks
            if supervisiontracks.count() > 0:
                now = datetime.datetime.now()
                for track in supervisiontracks:
                    track.is_valid = False
                    track.expiration_date = now
                    track.save()

        #Create and assign key
        supervisee.profile.supervision_key = supervisee.profile._generate_current_supervision_key()
        supervisiontrack = SupervisionTrack.objects.create(
            supervisee = supervisee,
            supervisor = supervisor,
            is_valid = False,
        )

        #Assign key to supervision track
        supervisiontrack.supervision_key = supervisee.profile.supervision_key
        supervisiontrack.save()
        supervisee.profile.supervision.add(supervisiontrack)
        supervisee.save()

        #Inform by email the supervisor that he has a new supervisee request
        context = {
            'supervisor': supervisor,
            'supervisee': supervisee,
            'prefix': server_address,
        }

        mail.send_email('registration/mail.supervisor_validation.subject.txt',
                        'registration/mail.supervisor_validation_supervisee_add_request.message.txt',
                        context,
                        [supervisor.email])

        return Response(status=status.HTTP_204_NO_CONTENT)


#----------------------------------------------------------


class BaseUpdateSupervisionTrackFromSuperviseeView(generics.UpdateAPIView):
    model = SupervisionTrack
    serializer_class = SupervisionTrackUpdateSerializer

    def get_permissions(self):
        permission_classes = [permissions.IsAuthenticated, IsAuthorAndPossiblySupervisor]

        self.permission_classes = permission_classes

        return super(BaseUpdateSupervisionTrackFromSuperviseeView, self).get_permissions()


#----------------------------------------------------------


class SuperviseeReValidationView(BaseUpdateSupervisionTrackFromSuperviseeView):
    permission_classes = BaseUpdateSupervisionTrackFromSuperviseeView.permission_classes

    def put(self, request):
        supervisee = request.user
        now = datetime.datetime.now()

        if supervisee.profile.supervision_key is not None:
            #There's a key check if there's a valid track
            supervisiontrack = SupervisionTrack.objects.get(supervisee=supervisee, supervisee__profile__supervision_key=models.F('supervision_key'))
            if supervisiontrack.is_valid:
                if supervisee.profile.status == Profile.YEARREVALIDATION:
                    #Check Supervisor validity
                    supervisor = supervisiontrack.supervisor
                    #If Supervisor account is not valid. Reject the account (though this should already be done during supervisor rejection)
                    if supervisor.profile.status != Profile.BLOCKED:
                        #Change status
                        supervisee.profile.status = Profile.ACCEPTED
                        #Extend supervisiontrack validity for another 12 months
                        expiration_date_delta = datetime.timedelta(days=settings.ACCOUNT_EXPIRATION_DAYS)
                        new_expiration_date = supervisiontrack.expiration_date + expiration_date_delta
                        supervisiontrack.expiration_date = new_expiration_date
                        supervisiontrack.last_validation_date = now
                    else:
                        #Change status
                        expiration_date_delta = datetime.timedelta(days=settings.ACCOUNT_BLOCKAGE_AFTER_FIRST_REJECTION_DAYS)

                        supervisiontrack.expiration_date = now
                        supervisiontrack.is_valid = False

                        supervisee.profile.status = Profile.REJECTED
                        supervisee.profile.rejection_date = now + expiration_date_delta
                        supervisee.profile.supervision_key = None
                    #save
                    supervisiontrack.save()
                    supervisee.profile.save()
                    supervisee.save()
                else:
                    #Track already valid
                    reason = "You don't need to revalidate at the moment, your supervision is still valid"
                    result = {
                        'error': reason,
                    }
                    return BadRequestResponse(result)
            else:
                #A pending request already exist
                reason = "You are not able to perform this action as you already have a pending supervision request"
                result = {
                    'error': reason,
                }
                return BadRequestResponse(result)
        else:
            #No key is present in supervisee
            #Make sure all tracks are invalid
            supervisiontracks = SupervisionTrack.objects.filter(supervisee=supervisee, is_valid=True)
            # This should never be the case but if it happens invalidate all tracks
            if supervisiontracks.count() > 0:
                now = datetime.datetime.now()
                for track in supervisiontracks:
                    track.is_valid = False
                    track.expiration_date = now
                    track.save()

            #Not allowed to do this (unproper profile.status)
            reason = "You are not allowed to perform this action, you first need to get a valid supervision"
            result = {
                'error': reason,
            }
            return BadRequestResponse(result)

        return Response(status=status.HTTP_204_NO_CONTENT)


#----------------------------------------------------------


class BaseUpdateSupervisorModeView(generics.UpdateAPIView):
    model = SupervisionTrack
    serializer_class = SupervisionTrackUpdateSerializer

    def get_permissions(self):
        permission_classes = [permissions.IsAuthenticated, IsSuperUser]

        self.permission_classes = permission_classes

        return super(BaseUpdateSupervisorModeView, self).get_permissions()



#----------------------------------------------------------


class SetSupervisorModeView(BaseUpdateSupervisorModeView):
    permission_classes = BaseUpdateSupervisorModeView.permission_classes

    def put(self, request):
        user = request.user
        now = datetime.datetime.now()

        if user.is_superuser:
            user.profile.is_supervisor = True
            user.profile.save()
            user.save()
        else:
            #Not allowed to do this (unproper profile.status - not admin)
            reason = "You are not allowed to perform this action"
            result = {
                'error': reason,
            }
            return BadRequestResponse(result)

        return Response(status=status.HTTP_204_NO_CONTENT)


#----------------------------------------------------------


class RemoveSupervisorModeView(BaseUpdateSupervisorModeView):
    permission_classes = BaseUpdateSupervisorModeView.permission_classes

    def put(self, request):
        user = request.user
        now = datetime.datetime.now()
        expiration_date_delta = datetime.timedelta(days=settings.ACCOUNT_BLOCKAGE_AFTER_FIRST_REJECTION_DAYS)

        if user.is_superuser:
            user.profile.is_supervisor = False
            user.profile.save()
            user.save()

            # Reject all his supervisees and inform by email
            # Make sure all supervision tracks are invalid for rejected supervisor
            supervisiontracks = SupervisionTrack.objects.filter(supervisor=user).filter(Q(supervisee__profile__status=Profile.REJECTED)|Q(supervisee__profile__status=Profile.YEARREVALIDATION)|Q(supervisee__profile__status=Profile.ACCEPTED)|Q(supervisee__profile__status=Profile.BLOCKED)).filter(Q(supervisee__profile__supervision_key=models.F('supervision_key'))).exclude(Q(supervisee__profile__is_supervisor=True))
            supervisiontracks_new_users = SupervisionTrack.objects.filter(supervisor=user).filter(Q(supervisee__profile__status=Profile.NEWUSER)|Q(supervisee__profile__status=Profile.WAITINGVALIDATION))

            parsed_url = urlparse(settings.URL_PREFIX)
            server_address = '%s://%s' % (parsed_url.scheme, parsed_url.hostname)

            if supervisiontracks.count() > 0:
                for track in supervisiontracks:
                    track.expiration_date = now
                    track.is_valid = False

                    track_supervisee = track.supervisee
                    if track_supervisee.profile.status != Profile.BLOCKED:
                        track_supervisee.profile.status = Profile.REJECTED
                        if track_supervisee.profile.rejection_date == None:
                            track_supervisee.profile.rejection_date = now + expiration_date_delta
                    else:
                            track_supervisee.profile.rejection_date = None

                    track_supervisee.profile.supervision_key = None

                    track.save()
                    track_supervisee.profile.save()
                    track_supervisee.save()

                    context = {
                        'supervisor': user,
                        'supervisee': track_supervisee,
                        'prefix': server_address,
                    }

                    mail.send_email('registration/mail.supervisor_failed_rejected.subject.txt',
                                    'registration/mail.supervisor_failed_rejected.message.txt',
                                    context,
                                    [track_supervisee.email])

                    if track_supervisee.profile.status == Profile.WAITINGVALIDATION:
                        registration_profile = RegistrationProfile.objects.get(user=track_supervisee)
                        track_supervisee.profile.delete()
                        track_supervisee.delete()
                        registration_profile.delete()

                    #New supervision request refused
                    if track.start_date is None:
                        track.delete()

            if supervisiontracks_new_users.count() > 0:
                for track in supervisiontracks_new_users:
                    track_supervisee = track.supervisee
                    track_supervisor = track.supervisor

                    context = {
                        'supervisor': track_supervisor,
                        'supervisee': track_supervisee,
                        'prefix': server_address,
                    }

                    #New user account waiting validation, so delete this account and inform by email the user
                    mail.send_email('registration/mail.supervisor_rejected.subject.txt',
                                    'registration/mail.supervisor_rejected_delete_account.message.txt',
                                    context,
                                    [track_supervisee.email])

                    registration_profile = RegistrationProfile.objects.get(user=track_supervisee)
                    track_supervisee.profile.delete()
                    track_supervisee.delete()
                    registration_profile.delete()

                    #New supervision request refused
                    if track.start_date is None:
                        track.delete()

        else:
            #Not allowed to do this (unproper profile.status - not admin)
            reason = "You are not allowed to perform this action"
            result = {
                'error': reason,
            }
            return BadRequestResponse(result)

        return Response(status=status.HTTP_204_NO_CONTENT)


#----------------------------------------------------------


class BaseListSupervisorCandidatesView(generics.ListAPIView):
    model = SupervisionTrack
    serializer_class = SupervisionTrackUpdateSerializer

    def get_permissions(self):
        permission_classes = [permissions.IsAuthenticated, IsSuperUser]

        self.permission_classes = permission_classes

        return super(BaseListSupervisorCandidatesView, self).get_permissions()



#----------------------------------------------------------


class ListSupervisorCandidatesView(BaseListSupervisorCandidatesView):
    permission_classes = BaseListSupervisorCandidatesView.permission_classes

    def list(self, request):
        user = request.user

        if user.is_superuser:
            specialusers = ["AnonymousUser", "plot", "system", "scheduler"]
            queryset    = User.objects.filter(profile__is_supervisor=False).filter(Q(profile__status=Profile.ACCEPTED)|Q(profile__status=Profile.REJECTED)|Q(profile__status=Profile.YEARREVALIDATION)|Q(profile__status=Profile.BLOCKED)).exclude(is_superuser=True).exclude(username__in=specialusers)
            serializer  = UserSerializer(queryset, many=True, context ={'request': request})
            return Response(serializer.data)

        else:
            #Not allowed to do this (unproper profile.status - not admin)
            reason = "You are not allowed to perform this action"
            result = {
                'error': reason,
            }
            return BadRequestResponse(result)


#----------------------------------------------------------


class BaseUpdateSupervisorCandidateView(generics.UpdateAPIView):
    model = SupervisionTrack
    serializer_class = SupervisionTrackUpdateSerializer

    def get_permissions(self):
        permission_classes = [permissions.IsAuthenticated, IsSuperUser]

        self.permission_classes = permission_classes

        return super(BaseUpdateSupervisorCandidateView, self).get_permissions()



#----------------------------------------------------------


class UpdateSupervisorCandidateView(BaseUpdateSupervisorCandidateView):
    permission_classes = BaseUpdateSupervisorCandidateView.permission_classes

    def put(self, request):
        user = request.user
        now = datetime.datetime.now()

        if user.is_superuser:
            try:
                supervisee_username = request.data['name']
            except:
                #Invalid supervisee username
                reason = "Invalid supervisee username"
                result = {
                    'error': reason,
                }
                return BadRequestResponse(result)

            supervisee = User.objects.get(username=supervisee_username)
            profile_supervisee = Profile.objects.get(user=supervisee)

            #Invalidate all valid tracks
            supervisiontracks = SupervisionTrack.objects.filter(supervisee=supervisee, is_valid=True)
            # This should never be the case but if it happens invalidate all tracks
            if supervisiontracks.count() > 0:
                for track in supervisiontracks:
                    track.is_valid = False
                    track.expiration_date = now
                    track.save()

            #Supervision from admin
            supervisee.profile.supervision_key = supervisee.profile._generate_current_supervision_key()

            supervisiontrack = SupervisionTrack.objects.create(
                supervisee = supervisee,
                supervisor = request.user,
                is_valid = True,
                )

            #Assign key to supervision track
            supervisiontrack.supervision_key = supervisee.profile.supervision_key
            supervisiontrack.save()

            supervisee.profile.supervision.add(supervisiontrack)
            supervisee.save()

            expiration_date_delta = datetime.timedelta(days=settings.ACCOUNT_EXPIRATION_DAYS)
            supervisiontrack.expiration_date = now + expiration_date_delta

            supervisiontrack.start_date = now
            supervisiontrack.last_validation_date = now
            supervisee.profile.status = Profile.ACCEPTED
            supervisee.profile.is_supervisor = True
            supervisee.profile.rejection_date = None

            supervisiontrack.save()
            supervisee.profile.save()
            supervisee.is_active = True
            supervisee.save()

            parsed_url = urlparse(settings.URL_PREFIX)
            server_address = '%s://%s' % (parsed_url.scheme, parsed_url.hostname)

            #Inform by email the new supervisor
            context = {
                'supervisor': request.user,
                'supervisee': supervisee,
                'prefix': server_address,
            }

            mail.send_email('registration/mail.grant_supervisor_access.subject.txt',
                            'registration/mail.grant_supervisor_access.message.txt',
                            context,
                            [supervisee.email])

            return Response(status=status.HTTP_204_NO_CONTENT)

        else:
            #Not allowed to do this (unproper profile.status - not admin)
            reason = "You are not allowed to perform this action"
            result = {
                'error': reason,
            }
            return BadRequestResponse(result)


#----------------------------------------------------------
