#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

###############################################################################
#                                                                             #
# Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.web module of the BEAT platform.              #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################

from django.contrib.auth.models import User

from ...libraries.models import Library
from ...common.testutils import tearDownModule  # noqa test runner will call it

from ..models import Algorithm

from .core import AlgorithmsBaseTestCase


class DependencyTests(AlgorithmsBaseTestCase):
    def setUp(self):
        super(DependencyTests, self).setUp()
        user = User.objects.get(username="johndoe")

        algorithm_declaration = """{
  "description": "some description",
  "schema_version": 1,
  "language": "python",
  "splittable": false,
  "groups": [
    {
      "name": "channel1",
      "inputs": {
        "in_data": {
          "type": "johndoe/integer/1"
        }
      },
      "outputs": {
        "out_data": {
          "type": "johndoe/integer/1"
        }
      }
    }
  ],
  "parameters": {
    "dummy": {
      "type": "int32",
      "default": "10"
    }
  }
}"""

        algorithm_code = """class Algorithm:

    def process(self, inputs, outputs):
        return True

"""

        library_declaration = """{
  "description": "some description",
  "language": "python"
}"""

        library_code = """class Library:

    def process(self, inputs, outputs):
        return True

"""
        (algorithm, errors) = Algorithm.objects.create_algorithm(
            author=user,
            name="algorithm1",
            declaration=algorithm_declaration,
            code=algorithm_code,
        )

        self.assertIsNotNone(algorithm, errors)

        (library, errors) = Library.objects.create_library(
            author=user,
            name="library1",
            declaration=library_declaration,
            code=library_code,
        )

        self.assertIsNotNone(library, errors)

    def test_add_dependency(self):
        algorithm1 = Algorithm.objects.get(name="algorithm1")
        library1 = Library.objects.get(name="library1")

        decl = algorithm1.declaration
        decl.setdefault("uses", {})
        decl["uses"]["dep1"] = library1.fullname()
        algorithm1.declaration = decl
        algorithm1.save()

        self.assertEqual(algorithm1.referenced_libraries.count(), 1)

    def test_update_dependency(self):
        algorithm1 = Algorithm.objects.get(name="algorithm1")
        library1 = Library.objects.get(name="library1")

        decl = algorithm1.declaration
        decl.setdefault("uses", {})
        decl["uses"]["dep1"] = library1.fullname()
        decl["uses"]["dep2"] = library1.fullname()
        algorithm1.declaration = decl
        algorithm1.save()

        self.assertEqual(algorithm1.referenced_libraries.count(), 1)

    def test_remove_dependency(self):
        self.test_update_dependency()
        algorithm1 = Algorithm.objects.get(name="algorithm1")
        self.assertEqual(algorithm1.referenced_libraries.count(), 1)

        decl = algorithm1.declaration
        decl["uses"] = {}
        algorithm1.declaration = decl
        algorithm1.save()

        self.assertEqual(algorithm1.referenced_libraries.count(), 0)
