#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

###############################################################################
#                                                                             #
# Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.web module of the BEAT platform.              #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################

import sys
import logging
import simplejson

# Default configuration to start the state with
import psutil
import socket

from django.core.management.base import BaseCommand

from ...utils import setup_backend, dump_backend

logger = logging.getLogger("beat.web")

CORES = psutil.cpu_count()
RAM = int(psutil.virtual_memory().total / (1024 * 1024))
ENVIRONMENT = {"name": "Python 2.7", "version": "1.3.0"}
CXX_ENVIRONMENT = {"name": "Cxx backend", "version": "1.1.0"}
GPU_ENVIRONMENT = {"name": "Pytorch 0.4.0", "version": "0.0.1"}
ENVKEY = "%(name)s (%(version)s)" % ENVIRONMENT
CXX_ENVKEY = "%(name)s (%(version)s)" % CXX_ENVIRONMENT
GPU_ENVKEY = "%(name)s (%(version)s)" % GPU_ENVIRONMENT
HOSTNAME = socket.gethostname()

DEFAULT_CONFIGURATION = {
    "queues": {
        "queue": {
            "memory-limit": int(RAM / CORES),
            "time-limit": 1440,  # 1 day
            "cores-per-slot": 1,
            "max-slots-per-user": CORES,
            "environments": [ENVKEY, CXX_ENVKEY, GPU_ENVKEY],
            "slots": {HOSTNAME: {"quantity": CORES, "priority": 0}},
            "groups": ["Default"],
        }
    },
    "environments": {
        ENVKEY: {
            "name": ENVIRONMENT["name"],
            "version": ENVIRONMENT["version"],
            "languages": ["python"],
            "short_description": "Local python interpreter",
            "description": "Automatically generated local python "
            "interpreter environment",
        },
        CXX_ENVKEY: {
            "name": CXX_ENVIRONMENT["name"],
            "version": CXX_ENVIRONMENT["version"],
            "languages": ["cxx"],
            "short_description": "C++ backend",
            "description": "C++ backend running in a docker container",
        },
        GPU_ENVKEY: {
            "name": GPU_ENVIRONMENT["name"],
            "version": GPU_ENVIRONMENT["version"],
            "languages": ["python"],
            "short_description": "Python GPU backend",
            "description": "Python backend running in a docker container with GPU support",
        },
    },
    "workers": {HOSTNAME: {"cores": CORES, "memory": RAM}},
}

RESET_CONFIGURATION = {"queues": {}, "environments": {}, "workers": {}}


class Command(BaseCommand):

    help = "Sets and resets queue configurations"

    def add_arguments(self, parser):

        parser.add_argument(
            "--dump",
            action="store_true",
            dest="dump",
            default=False,
            help="Dump all environment/worker/queues " "configuration and exits",
        )

        parser.add_argument(
            "--reset",
            action="store_true",
            dest="reset",
            default=False,
            help="Delete all environment/worker/queues "
            "before setting the given configuration",
        )

        parser.add_argument(
            "config",
            type=str,
            nargs="?",
            help="Optional custom queue configuration to use. If not "
            "passed, uses an internal default with a single "
            "queue/worker/%d slots" % CORES,
        )

    def handle(self, *ignored, **arguments):

        # Setup this command's logging level
        global logger
        arguments["verbosity"] = int(arguments["verbosity"])
        if arguments["verbosity"] >= 1:
            if arguments["verbosity"] == 1:
                logger.setLevel(logging.INFO)
            elif arguments["verbosity"] >= 2:
                logger.setLevel(logging.DEBUG)

        if arguments["dump"]:
            d = dump_backend()
            print(simplejson.dumps(d, indent=2))
            sys.exit(0)

        if arguments["reset"]:
            setup_backend(RESET_CONFIGURATION)

        config = None
        if arguments["config"]:
            with open(arguments["config"], "rb") as f:
                config = simplejson.load(f)

        setup_backend(config or DEFAULT_CONFIGURATION)
