#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

###############################################################################
#                                                                             #
# Copyright (c) 2017 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.web module of the BEAT platform.              #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################


import operator

from django.db import models
from django.core.urlresolvers import reverse
from django.contrib.auth.models import Group
from django.utils.translation import ugettext_lazy as _

from guardian.shortcuts import get_perms

from ...common.texts import Messages


#----------------------------------------------------------


class QueueManager(models.Manager):

    def get_by_natural_key(self, name):
        return self.get(name=name)


#----------------------------------------------------------


class Queue(models.Model):

    name = models.CharField(max_length=100, help_text=Messages['name'],
                            unique=True)

    memory_limit = models.PositiveIntegerField(help_text='In megabytes')

    time_limit = models.PositiveIntegerField(help_text='In minutes')

    cores_per_slot = models.PositiveIntegerField()

    max_slots_per_user = models.PositiveIntegerField()

    environments = models.ManyToManyField('Environment', related_name='queues')


    objects = QueueManager()


    #_____ Meta parameters __________

    class Meta:
        permissions = [
            ['can_access', _('Can access queue')]
        ]


    #_____ Overrides __________

    def __str__(self):
        return '%s (%d minutes, %d megabytes, %d cores per slot, %d slots max per user)' % (
            self.name,
            self.time_limit,
            self.memory_limit,
            self.cores_per_slot,
            self.max_slots_per_user
        )


    def natural_key(self):
        return (self.name,)


    def get_admin_change_url(self):
        return reverse('admin:backend_queue_change', args=(self.id,))


    #_____ Utilities __________

    def number_of_slots(self):
        '''Total number of slots considering all assigned worker/slots'''

        r = self.slots.filter(worker__active=True)
        return r.aggregate(nslots=models.Sum('quantity'))['nslots'] or 0


    def availability(self):
        '''Returns the availability for this queue in terms of number of slots

        This method does not take into consideration the occupation of this
        queue slots caused by jobs on other queues. It only looks to its inner
        occupancy and reports on that.

        Returns an integer between 0 and :py:meth:`Queue.slots`.
        '''

        from ..models import JobSplit
        from ..models import Job

        running = JobSplit.objects.filter(job__block__in=self.blocks.all(),
                                          status=JobSplit.PROCESSING).count()
        return max(self.number_of_slots() - running, 0)


    def worker_availability(self):
        '''Returns an ordered dictionary indicating the availability of workers
           according to their queue priority.

        The dictionary contains, as value, the number of slots available per
        worker

        The order of workers is sorted by:

          * slot priority (the higher, the better)
          * load (the lower, the better)
          * name (alphabetically)

        '''

        workers = [(k.worker, -k.priority, k.worker.load(), k.worker.name) \
                   for k in self.slots.filter(worker__active=True)]

        workers = sorted(workers, key=operator.itemgetter(1,2,3))

        return [w[0] for w in workers]


    def splits(self):
        '''Lists all job splits currently associated to this queue'''
        from ..models import JobSplit
        from ..models import Job

        return JobSplit.objects.filter(job__block__queue=self)


    def as_dict(self):
        '''Returns a representation as a dictionary'''

        return {
            'memory-limit': self.memory_limit,
            'time-limit': self.time_limit,
            'cores-per-slot': self.cores_per_slot,
            'max-slots-per-user': self.max_slots_per_user,
            'environments': [k.fullname() for k in self.environments.all()],
            'slots': dict([(s.worker.name, dict(quantity=s.quantity, priority=s.priority))
                           for s in self.slots.all()]),
            'groups': [k.name for k in Group.objects.all() if 'can_access' in get_perms(k, self)]
        }
