#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

###############################################################################
#                                                                             #
# Copyright (c) 2017 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.web module of the BEAT platform.              #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################

import os
import time
import shutil
import tempfile

from django.core import management
from django.test import TestCase

from ...common.testutils import tearDownModule  # noqa test runner will call it
from ..utils import cleanup_cache


class CacheCleanUp(TestCase):
    def setUp(self):
        self.cache = tempfile.mkdtemp(prefix="beat_")

    def tearDown(self):
        shutil.rmtree(self.cache)

    def touch(self, f, times=None):
        """Replicates the `touch' command-line utility"""
        with open(f, "a"):
            os.utime(f, times)

    def J(self, *args):
        return os.path.join(*((self.cache,) + args))

    def prepare_cleanup_full(self):

        # creates a temporary directory structure
        os.makedirs(self.J("a", "b", "c"))
        os.makedirs(self.J("a", "c", "d"))
        os.makedirs(self.J("a", "c", "e"))
        self.touch(self.J("a", "b", "c", "d.json"))
        self.touch(self.J("a", "c", "d", "e.json"))

    def check_cleanup_full(self):
        self.assertIsNotNone(os.listdir(self.cache))

    def test_cache_cleanup_full(self):

        self.prepare_cleanup_full()
        cleanup_cache(self.cache, delete=True)
        self.check_cleanup_full()

    def test_cmd_cleanup_full(self):

        self.prepare_cleanup_full()
        management.call_command(
            "cleanup_cache", path=self.cache, verbosity=0, delete=True
        )
        self.check_cleanup_full()

    def prepare_cleanup_aged(self):

        two_min_ago = time.time() - 60 * 2

        # creates a temporary directory structure
        os.makedirs(self.J("a", "b", "c"))
        os.makedirs(self.J("a", "c", "d"))
        os.makedirs(self.J("a", "c", "e"))
        self.touch(self.J("a", "b", "c", "d.json"), (two_min_ago, two_min_ago))
        self.touch(self.J("a", "c", "d", "e.json"))

    def check_cleanup_aged(self):

        self.assertTrue(os.path.exists(self.J("a", "c", "d", "e.json")))
        self.assertFalse(os.path.exists(self.J("a", "b", "c")))
        self.assertFalse(os.path.exists(self.J("a", "b", "c", "d.json")))
        self.assertFalse(os.path.exists(self.J("a", "b", "e")))

    def test_cache_cleanup_aged(self):

        self.prepare_cleanup_aged()
        cleanup_cache(self.cache, age_in_minutes=2, delete=True)
        self.check_cleanup_aged()

    def test_cmd_cleanup_aged(self):

        self.prepare_cleanup_aged()
        management.call_command(
            "cleanup_cache", path=self.cache, verbosity=0, olderthan=2, delete=True
        )
        self.check_cleanup_aged()

    def prepare_cleanup_lock(self):

        two_min_ago = time.time() - 60 * 2
        ten_min_ago = time.time() - 60 * 10

        # creates a temporary directory structure
        os.makedirs(self.J("a", "b", "c"))
        os.makedirs(self.J("a", "c", "d"))
        os.makedirs(self.J("a", "c", "e"))
        self.touch(self.J("a", "b", "c", "d.json"), (two_min_ago, two_min_ago))
        self.touch(self.J("a", "c", "d", "e.json"), (ten_min_ago, ten_min_ago))

        self.touch(self.J("a", "c", "d", "e.lock"))  # create a lock

    def check_cleanup_lock(self):

        self.assertTrue(os.path.exists(self.J("a", "c", "d", "e.json")))
        self.assertFalse(os.path.exists(self.J("a", "b", "c")))
        self.assertFalse(os.path.exists(self.J("a", "b", "c", "d.json")))
        self.assertFalse(os.path.exists(self.J("a", "b", "e")))

    def test_cache_cleanup_lock(self):

        self.prepare_cleanup_lock()
        cleanup_cache(self.cache, delete=True)
        self.check_cleanup_lock()

    def test_cmd_cleanup_lock(self):

        self.prepare_cleanup_lock()
        management.call_command(
            "cleanup_cache", path=self.cache, verbosity=0, delete=True
        )
        self.check_cleanup_lock()
