#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

###############################################################################
#                                                                             #
# Copyright (c) 2017 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.web module of the BEAT platform.              #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################

import os

from time import time
from time import sleep
from unittest import skipIf

from django.conf import settings
from django.test import TransactionTestCase

from ...experiments.models import Experiment
from ...experiments.models import Block
from ...common.testutils import tearDownModule  # noqa test runner will call it

from ..models import Worker

from ..utils import setup_backend
from ..helpers import schedule_experiment
from ..helpers import cancel_experiment
from ..local_scheduler import start_scheduler
from ..local_scheduler import start_worker

from .common import BackendUtilitiesMixin
from .common import ONE_QUEUE_TWO_WORKERS


# ----------------------------------------------------------


class TestSchedulerBase(TransactionTestCase, BackendUtilitiesMixin):
    def __init__(self, methodName="runTest"):
        super(TestSchedulerBase, self).__init__(methodName)
        self.scheduler_thread = None
        self.worker_threads = {}
        self.timeout = int(os.environ.get("DB_REFRESH_TIMEOUT", 10))

    @classmethod
    def setUpTestData(cls):
        cls.setup_test_data()

    def setUp(self):
        self.shutdown_everything()
        self.clean_cache()

        TestSchedulerBase.setup_test_data()
        setup_backend(ONE_QUEUE_TWO_WORKERS)
        Worker.objects.update(active=False)

    def tearDown(self):
        self.shutdown_everything()
        self.clean_cache()

    def shutdown_everything(self):
        for name in list(self.worker_threads.keys()):
            self.stop_worker(name)

        self.worker_threads = {}

        self.stop_scheduler()

    def start_scheduler(self):
        (pid, self.scheduler_thread) = start_scheduler(
            settings_module="beat.web.settings.test",
            interval=1,
            address="127.0.0.1",
            port=50800,
        )

    def stop_scheduler(self):
        if self.scheduler_thread is not None:
            self.scheduler_thread.terminate()
            self.scheduler_thread.join()
            self.scheduler_thread = None

    def start_worker(self, name):
        (pid, worker_thread) = start_worker(
            name, settings.PREFIX, settings.CACHE_ROOT, "tcp://127.0.0.1:50800"
        )

        self.worker_threads[name] = worker_thread

    def stop_worker(self, name):
        if name in self.worker_threads:
            self.worker_threads[name].terminate()
            self.worker_threads[name].join()
            del self.worker_threads[name]

    def check_worker_status(self, name, active):
        start = time()
        while Worker.objects.filter(name=name, active=active).count() == 0:
            if time() - start > 10:  # Fail after 10 seconds
                self.shutdown_everything()
                self.assertTrue(False)


# ----------------------------------------------------------


@skipIf(settings.RUNNING_ON_CI, "Not runnable on the CI")
class TestConnection(TestSchedulerBase):
    def test_worker_connection(self):
        self.start_scheduler()

        self.assertEqual(Worker.objects.filter(active=True).count(), 0)

        self.start_worker("node1")

        self.check_worker_status("node1", True)

        self.assertEqual(Worker.objects.filter(active=True).count(), 1)

    def test_worker_disconnection(self):
        self.start_scheduler()
        self.start_worker("node1")

        self.check_worker_status("node1", True)

        self.worker_threads["node1"].terminate()
        self.worker_threads["node1"].join()
        del self.worker_threads["node1"]

        self.check_worker_status("node1", False)

    def test_two_workers_connection(self):
        self.start_scheduler()

        self.assertEqual(Worker.objects.filter(active=True).count(), 0)

        self.start_worker("node1")
        self.start_worker("node2")

        self.check_worker_status("node1", True)
        self.check_worker_status("node2", True)

        self.assertEqual(Worker.objects.filter(active=True).count(), 2)

    def test_scheduler_last(self):
        self.start_worker("node1")

        sleep(1)

        self.start_scheduler()

        self.check_worker_status("node1", True)

    def test_scheduler_shutdown(self):
        self.start_scheduler()

        self.assertEqual(Worker.objects.filter(active=True).count(), 0)

        self.start_worker("node1")

        self.check_worker_status("node1", True)

        self.assertEqual(Worker.objects.filter(active=True).count(), 1)

        self.stop_scheduler()

        Worker.objects.update(active=False)

        sleep(1)

        self.start_scheduler()

        self.check_worker_status("node1", True)

        self.assertEqual(Worker.objects.filter(active=True).count(), 1)

    def test_worker_disconnection_during_scheduler_shutdown(self):
        self.start_scheduler()

        self.assertEqual(Worker.objects.filter(active=True).count(), 0)

        self.start_worker("node1")

        self.check_worker_status("node1", True)

        self.assertEqual(Worker.objects.filter(active=True).count(), 1)

        self.stop_scheduler()

        Worker.objects.update(active=False)

        sleep(1)

        self.worker_threads["node1"].terminate()
        self.worker_threads["node1"].join()
        del self.worker_threads["node1"]

        sleep(1)

        self.start_scheduler()

        sleep(5)

        self.check_worker_status("node1", False)

        self.assertEqual(Worker.objects.filter(active=True).count(), 0)


# ----------------------------------------------------------


@skipIf(settings.RUNNING_ON_CI, "Not runnable on the CI")
class TestExecution(TestSchedulerBase):
    def setUp(self):
        super(TestExecution, self).setUp()

        self.start_scheduler()
        self.start_worker("node1")
        self.start_worker("node2")
        self.check_worker_status("node1", True)
        self.check_worker_status("node2", True)

    def test_successful_experiment(self):
        fullname = "user/user/double/1/double"

        xp = Experiment.objects.get(name=fullname.split("/")[-1])

        self.prepare_databases(xp.declaration)

        schedule_experiment(xp)
        xp.refresh_from_db()

        start = time()
        while xp.status != Experiment.DONE:
            self.assertTrue(time() - start < self.timeout)
            xp.refresh_from_db()

        b1 = xp.blocks.get(name="echo1")
        b2 = xp.blocks.get(name="echo2")
        b3 = xp.blocks.get(name="analysis")

        self.assertEqual(b1.status, Block.DONE)
        self.assertEqual(b2.status, Block.DONE)
        self.assertEqual(b3.status, Block.DONE)

    def test_successful_splitted_experiment(self):
        fullname = "user/user/double/1/double_split_2"

        xp = Experiment.objects.get(name=fullname.split("/")[-1])

        self.prepare_databases(xp.declaration)

        schedule_experiment(xp)
        xp.refresh_from_db()

        start = time()
        while xp.status != Experiment.DONE:
            self.assertTrue(time() - start < self.timeout)
            xp.refresh_from_db()

        b1 = xp.blocks.get(name="echo1")
        b2 = xp.blocks.get(name="echo2")
        b3 = xp.blocks.get(name="analysis")

        self.assertEqual(b1.status, Block.DONE)
        self.assertEqual(b2.status, Block.DONE)
        self.assertEqual(b3.status, Block.DONE)

    def test_two_similar_experiments(self):
        fullname1 = "user/user/double/1/double"
        fullname2 = "user/user/double/1/double_split_2"

        xp1 = Experiment.objects.get(name=fullname1.split("/")[-1])
        xp2 = Experiment.objects.get(name=fullname2.split("/")[-1])

        self.prepare_databases(xp1.declaration)
        self.prepare_databases(xp2.declaration)

        schedule_experiment(xp1)
        schedule_experiment(xp2)

        xp1.refresh_from_db()
        xp2.refresh_from_db()

        start = time()
        while (xp1.status != Experiment.DONE) or (xp2.status != Experiment.DONE):
            self.assertTrue(time() - start < self.timeout)
            xp1.refresh_from_db()
            xp2.refresh_from_db()

        self.assertEqual(xp1.start_date, xp2.start_date)
        self.assertEqual(xp1.end_date, xp2.end_date)

        b1 = xp1.blocks.get(name="echo1")
        b2 = xp2.blocks.get(name="echo1")

        self.assertEqual(b1.status, Block.DONE)
        self.assertEqual(b2.status, Block.DONE)
        self.assertEqual(b1.start_date, b2.start_date)
        self.assertEqual(b1.end_date, b2.end_date)

        b1 = xp1.blocks.get(name="echo2")
        b2 = xp2.blocks.get(name="echo2")

        self.assertEqual(b1.status, Block.DONE)
        self.assertEqual(b2.status, Block.DONE)
        self.assertEqual(b1.start_date, b2.start_date)
        self.assertEqual(b1.end_date, b2.end_date)

        b1 = xp1.blocks.get(name="analysis")
        b2 = xp2.blocks.get(name="analysis")

        self.assertEqual(b1.status, Block.DONE)
        self.assertEqual(b2.status, Block.DONE)
        self.assertEqual(b1.start_date, b2.start_date)
        self.assertEqual(b1.end_date, b2.end_date)

    def test_two_different_experiments(self):
        fullname1 = "user/user/single/1/single"
        fullname2 = "user/user/single/1/single_add"

        xp1 = Experiment.objects.get(name=fullname1.split("/")[-1])
        xp2 = Experiment.objects.get(name=fullname2.split("/")[-1])

        self.prepare_databases(xp1.declaration)
        self.prepare_databases(xp2.declaration)

        schedule_experiment(xp1)
        schedule_experiment(xp2)

        xp1.refresh_from_db()
        xp2.refresh_from_db()

        start = time()
        while (xp1.status != Experiment.DONE) or (xp2.status != Experiment.DONE):
            self.assertTrue(time() - start < self.timeout)
            xp1.refresh_from_db()
            xp2.refresh_from_db()

        self.assertNotEqual(xp1.start_date, xp2.start_date)
        self.assertNotEqual(xp1.end_date, xp2.end_date)

    def test_two_delayed_similar_experiments(self):
        fullname1 = "user/user/double/1/double"
        fullname2 = "user/user/double/1/double_split_2"

        xp1 = Experiment.objects.get(name=fullname1.split("/")[-1])
        xp2 = Experiment.objects.get(name=fullname2.split("/")[-1])

        self.prepare_databases(xp1.declaration)
        self.prepare_databases(xp2.declaration)

        schedule_experiment(xp1)

        xp1.refresh_from_db()

        start = time()
        while xp1.status != Experiment.RUNNING:
            self.assertTrue(time() - start < self.timeout)
            xp1.refresh_from_db()

        schedule_experiment(xp2)
        xp2.refresh_from_db()

        start = time()
        while (xp1.status != Experiment.DONE) or (xp2.status != Experiment.DONE):
            self.assertTrue(time() - start < self.timeout)
            xp1.refresh_from_db()
            xp2.refresh_from_db()

    def test_failed_experiment(self):
        fullname = "user/user/single/1/single_error"

        xp = Experiment.objects.get(name=fullname.split("/")[-1])

        self.prepare_databases(xp.declaration)

        schedule_experiment(xp)
        xp.refresh_from_db()

        start = time()
        while xp.status != Experiment.FAILED:
            self.assertTrue(time() - start < self.timeout)
            xp.refresh_from_db()

        b1 = xp.blocks.get(name="echo")
        b2 = xp.blocks.get(name="analysis")

        self.assertEqual(b1.status, Block.FAILED)
        self.assertEqual(b2.status, Block.CANCELLED)

    def test_failed_splitted_experiment(self):
        fullname = "user/user/double/1/double_error_split_2"

        xp = Experiment.objects.get(name=fullname.split("/")[-1])

        self.prepare_databases(xp.declaration)

        schedule_experiment(xp)
        xp.refresh_from_db()

        start = time()
        while xp.status != Experiment.FAILED:
            self.assertTrue(time() - start < self.timeout)
            xp.refresh_from_db()

        b1 = xp.blocks.get(name="echo1")
        b2 = xp.blocks.get(name="echo2")
        b3 = xp.blocks.get(name="analysis")

        self.assertEqual(b1.status, Block.DONE)
        self.assertEqual(b2.status, Block.FAILED)
        self.assertEqual(b3.status, Block.CANCELLED)

    def test_failed_mirror_experiments(self):
        fullname1 = "user/user/double/1/double_error"
        fullname2 = "user/user/double/1/double_error_split_2"

        xp1 = Experiment.objects.get(name=fullname1.split("/")[-1])
        xp2 = Experiment.objects.get(name=fullname2.split("/")[-1])

        self.prepare_databases(xp1.declaration)
        self.prepare_databases(xp2.declaration)

        schedule_experiment(xp1)
        schedule_experiment(xp2)

        xp1.refresh_from_db()
        xp2.refresh_from_db()

        start = time()
        while (xp1.status != Experiment.FAILED) or (xp2.status != Experiment.FAILED):
            self.assertTrue(time() - start < self.timeout)
            xp1.refresh_from_db()
            xp2.refresh_from_db()

        b1 = xp1.blocks.get(name="echo1")
        b2 = xp1.blocks.get(name="echo2")
        b3 = xp1.blocks.get(name="analysis")

        self.assertEqual(b1.status, Block.DONE)
        self.assertEqual(b2.status, Block.FAILED)
        self.assertEqual(b3.status, Block.CANCELLED)

        b1 = xp2.blocks.get(name="echo1")
        b2 = xp2.blocks.get(name="echo2")
        b3 = xp2.blocks.get(name="analysis")

        self.assertEqual(b1.status, Block.DONE)
        self.assertEqual(b2.status, Block.FAILED)
        self.assertEqual(b3.status, Block.CANCELLED)


# ----------------------------------------------------------


@skipIf(settings.RUNNING_ON_CI, "Not runnable on the CI")
class TestCancellation(TestSchedulerBase):
    def setUp(self):
        super(TestCancellation, self).setUp()

        self.start_scheduler()
        self.start_worker("node1")
        self.start_worker("node2")
        self.check_worker_status("node1", True)
        self.check_worker_status("node2", True)

    def process(self, experiment_name, block_name=None):
        xp = Experiment.objects.get(name=experiment_name.split("/")[-1])

        self.prepare_databases(xp.declaration)

        schedule_experiment(xp)
        xp.refresh_from_db()

        start = time()
        while xp.status != Experiment.RUNNING:
            self.assertTrue(time() - start < self.timeout)
            xp.refresh_from_db()

        if block_name is not None:
            block = xp.blocks.get(name=block_name)

            start = time()
            while block.status != Block.PROCESSING:
                self.assertTrue(time() - start < self.timeout)
                block.refresh_from_db()

        cancel_experiment(xp)
        xp.refresh_from_db()

        start = time()
        while xp.status != Experiment.PENDING:
            self.assertTrue(time() - start < self.timeout)
            xp.refresh_from_db()

        self.assertTrue(xp.blocks.filter(status=Block.CANCELLED).count() > 0)
        self.assertEqual(xp.blocks.filter(status=Block.PENDING).count(), 0)
        self.assertEqual(xp.blocks.filter(status=Block.PROCESSING).count(), 0)

    def process2(self, experiment_name1, experiment_name2, cancel_index=0):
        xp1 = Experiment.objects.get(name=experiment_name1.split("/")[-1])
        xp2 = Experiment.objects.get(name=experiment_name2.split("/")[-1])

        self.prepare_databases(xp1.declaration)
        self.prepare_databases(xp2.declaration)

        schedule_experiment(xp1)
        schedule_experiment(xp2)
        xp1.refresh_from_db()
        xp2.refresh_from_db()

        start = time()
        while (xp1.status != Experiment.RUNNING) or (xp2.status != Experiment.RUNNING):
            self.assertTrue(time() - start < self.timeout)
            xp1.refresh_from_db()
            xp2.refresh_from_db()

        if cancel_index == 0:
            xp_to_cancel = xp1
            xp_to_finish = xp2
        else:
            xp_to_cancel = xp2
            xp_to_finish = xp1

        cancel_experiment(xp_to_cancel)

        xp_to_cancel.refresh_from_db()
        xp_to_finish.refresh_from_db()

        start = time()
        while xp_to_cancel.status != Experiment.PENDING:
            self.assertTrue(time() - start < self.timeout)
            xp_to_cancel.refresh_from_db()

        xp_to_finish.refresh_from_db()

        start = time()
        while xp_to_finish.status != Experiment.DONE:
            self.assertTrue(time() - start < self.timeout)
            xp_to_finish.refresh_from_db()

        self.assertTrue(xp_to_cancel.blocks.filter(status=Block.CANCELLED).count() > 0)
        self.assertEqual(xp_to_cancel.blocks.filter(status=Block.PENDING).count(), 0)
        self.assertEqual(xp_to_cancel.blocks.filter(status=Block.PROCESSING).count(), 0)

        self.assertEqual(xp_to_finish.blocks.filter(status=Block.CANCELLED).count(), 0)
        self.assertEqual(xp_to_finish.blocks.filter(status=Block.PENDING).count(), 0)
        self.assertEqual(xp_to_finish.blocks.filter(status=Block.PROCESSING).count(), 0)

    def test_one_split_running(self):
        self.process("user/user/single/1/single_sleep_4")

    def test_one_split_of_two_in_a_block_running(self):
        self.stop_worker("node2")
        self.process("user/user/double/1/double_sleep_split_2", block_name="echo2")

    def test_two_splits_of_same_block_running(self):
        self.process("user/user/double/1/double_sleep_split_2", block_name="echo2")

    def test_two_splits_of_different_blocks_running(self):
        self.process("user/user/triangle/1/triangle_sleep_4")

    def test_mirror_block(self):
        self.stop_worker("node2")
        self.process2(
            "user/user/single/1/single_sleep_4",
            "user/user/single/1/single_sleep_5",
            cancel_index=1,
        )

    def test_mirrored_block(self):
        self.stop_worker("node2")
        self.process2(
            "user/user/single/1/single_sleep_4",
            "user/user/single/1/single_sleep_5",
            cancel_index=0,
        )
