#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

###############################################################################
#                                                                             #
# Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.web module of the BEAT platform.              #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################


import os
import shutil

from django.conf import settings

from django.contrib.auth.models import User

from ..models import DataFormat

from ...team.models import Team
from ...common.testutils import BaseTestCase
from ...common.testutils import tearDownModule  # noqa test runner will call it

TEST_PWD = "1234"


class DataFormatsAccessibilityFunctionsBase(BaseTestCase):
    def setUp(self):
        if os.path.exists(settings.DATAFORMATS_ROOT):
            shutil.rmtree(settings.DATAFORMATS_ROOT)

        # Users
        user1 = User.objects.create_user("jackdoe", "jackdoe@test.org", TEST_PWD)
        user2 = User.objects.create_user("johndoe", "johndoe@test.org", TEST_PWD)
        user3 = User.objects.create_user("janedoe", "janedoe@test.org", TEST_PWD)
        User.objects.create_user("jacquelinedoe", "jacquelinedoe@test.org", TEST_PWD)

        team1 = Team.objects.create(name="teamdoe", owner=user1)
        team2 = Team.objects.create(name="teamdoe2", owner=user1)
        Team.objects.create(name="teamdoe3", owner=user1)
        Team.objects.create(name="teamdoe4", owner=user1)

        team1.members.add(user2)
        team2.members.add(user3)

        # Personal dataformat not shared with anyone
        (dataformat, errors) = DataFormat.objects.create_dataformat(
            author=user1, name="personal"
        )
        self.assertIsNotNone(dataformat, errors)

        # Accessible to one user
        (dataformat, errors) = DataFormat.objects.create_dataformat(
            author=user1, name="public_for_one_user"
        )
        self.assertIsNotNone(dataformat, errors)

        dataformat.share(users=[user2])

        # Accessible to one team
        (dataformat, errors) = DataFormat.objects.create_dataformat(
            author=user1, name="public_for_one_team"
        )
        self.assertIsNotNone(dataformat, errors)

        dataformat.share(teams=[team1])

        # Accessible to teams
        (dataformat, errors) = DataFormat.objects.create_dataformat(
            author=user1, name="public_for_teams"
        )
        self.assertIsNotNone(dataformat, errors)

        dataformat.share(teams=[team1, team2])

        # Accessible to all
        (dataformat, errors) = DataFormat.objects.create_dataformat(
            author=user1, name="public_for_all"
        )
        self.assertIsNotNone(dataformat, errors)

        dataformat.share()

    def tearDown(self):
        if os.path.exists(settings.DATAFORMATS_ROOT):
            shutil.rmtree(settings.DATAFORMATS_ROOT)


class DataFormatsAPIBase(BaseTestCase):
    class Meta:
        model = DataFormat

    def setUp(self):
        if os.path.exists(settings.DATAFORMATS_ROOT):
            shutil.rmtree(settings.DATAFORMATS_ROOT)

        user1 = User.objects.create_user("johndoe", "johndoe@test.org", TEST_PWD)
        user2 = User.objects.create_user("jackdoe", "jackdoe@test.org", TEST_PWD)
        User.objects.create_user("janedoe", "janedoe@test.org", TEST_PWD)
        user4 = User.objects.create_user(
            "jacquelinedoe", "jacquelinedoe@test.org", TEST_PWD
        )

        team1 = Team.objects.create(name="teamdoe", owner=user2)
        Team.objects.create(name="teamdoe2", owner=user2)
        team1.members.add(user4)

        (dataformat_private, errors) = DataFormat.objects.create_dataformat(
            author=user1,
            name="format_private",
            short_description="short private description",
            description="long private description",
        )
        self.assertIsNotNone(dataformat_private, errors)

        (dataformat1, errors) = DataFormat.objects.create_dataformat(
            author=user1,
            name="format1",
            short_description="short description 1",
            description="long description 1",
        )
        self.assertIsNotNone(dataformat1, errors)
        dataformat1.share()

        (dataformat2, errors) = DataFormat.objects.create_dataformat(
            author=user2,
            name="format2",
            short_description="short description 2",
            description="long description 2",
        )
        self.assertIsNotNone(dataformat2, errors)

        (dataformat3, errors) = DataFormat.objects.create_dataformat(
            author=user2,
            name="format3",
            short_description="short description 3",
            description="long description 3",
            declaration='{\n  "#extends": "%s"\n}' % dataformat1.fullname(),
        )
        self.assertIsNotNone(dataformat3, errors)
        dataformat3.share()

        (dataformat4, errors) = DataFormat.objects.create_dataformat(
            author=user2,
            name="team_format1",
            short_description="team short description 1",
            description="team long description 1",
        )
        self.assertIsNotNone(dataformat4, errors)
        dataformat4.share(teams=[team1])

    def tearDown(self):
        if os.path.exists(settings.DATAFORMATS_ROOT):
            shutil.rmtree(settings.DATAFORMATS_ROOT)

    def login(self, username):
        self.client.login(username=username, password=TEST_PWD)

    def login_johndoe(self):
        self.login("johndoe")

    def login_jackdoe(self):
        self.login("jackdoe")


class DataFormatSharingAPIBase(DataFormatsAPIBase):
    def setUp(self):
        super(DataFormatSharingAPIBase, self).setUp()

        # Users
        user1 = User.objects.get(username="jackdoe")
        user2 = User.objects.get(username="johndoe")
        User.objects.get(username="janedoe")
        User.objects.get(username="jacquelinedoe")

        team1 = Team.objects.get(name="teamdoe")
        team2 = Team.objects.get(name="teamdoe2")
        Team.objects.create(name="teamdoe3", owner=user1)
        Team.objects.create(name="teamdoe4", owner=user1)

        # Personal dataformat not shared with anyone
        (dataformat, errors) = DataFormat.objects.create_dataformat(
            author=user1, name="personal"
        )
        self.assertIsNotNone(dataformat, errors)

        # Accessible to one user
        (dataformat, errors) = DataFormat.objects.create_dataformat(
            author=user1, name="public_for_one_user"
        )
        self.assertIsNotNone(dataformat, errors)
        dataformat.share(users=[user2])

        # Accessible to one team
        (dataformat, errors) = DataFormat.objects.create_dataformat(
            author=user1, name="public_for_one_team"
        )
        self.assertIsNotNone(dataformat, errors)
        dataformat.share(teams=[team1])

        # Accessible to teams
        (dataformat, errors) = DataFormat.objects.create_dataformat(
            author=user1, name="public_for_teams"
        )
        self.assertIsNotNone(dataformat, errors)
        dataformat.share(teams=[team1, team2])

        # Accessible to all
        (dataformat, errors) = DataFormat.objects.create_dataformat(
            author=user1, name="public_for_all"
        )
        self.assertIsNotNone(dataformat, errors)
        dataformat.share()

        # Personal dataformat not shared with anyone, extension of a personal one
        (dataformat, errors) = DataFormat.objects.create_dataformat(
            author=user1,
            name="extension_of_personal",
            declaration='{\n  "#extends": "%s/personal/1"\n}' % user1.username,
        )
        self.assertIsNotNone(dataformat, errors)

        # Personal dataformat not shared with anyone, extension of a public one
        (dataformat, errors) = DataFormat.objects.create_dataformat(
            author=user1,
            name="extension_of_public",
            declaration='{\n  "#extends": "%s/public_for_all/1"\n}' % user1.username,
        )
        self.assertIsNotNone(dataformat, errors)

        # Personal dataformat not shared with anyone, extension of a limited public one
        (dataformat, errors) = DataFormat.objects.create_dataformat(
            author=user2,
            name="extension_of_external",
            declaration='{\n  "#extends": "%s/public_for_one_user/1"\n}'
            % user1.username,
        )
        self.assertIsNotNone(dataformat, errors)

    def checkJSONSharingPreferences(self, sharing_preferences, reference):
        self.assertEqual(len(sharing_preferences.keys()), len(reference.keys()))

        for (key, value) in reference.items():
            self.assertTrue(key in sharing_preferences)

            if isinstance(value, list):
                self.assertEqual(len(sharing_preferences[key]), len(value))
                for entry in value:
                    self.assertTrue(entry in sharing_preferences[key])
            else:
                self.assertEqual(sharing_preferences[key], value)
