#!/usr/bin/env python
# vim: set fileencoding=utf-8 :
# encoding: utf-8

###############################################################################
#                                                                             #
# Copyright (c) 2019 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.web module of the BEAT platform.              #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################

import logging

from django.core.management.base import BaseCommand
from django.db import transaction

from beat.web.backend.models import Worker
from beat.core.bcpapi.broker import BeatComputationBroker
from beat.core.utils import setup_logging

logger = logging.getLogger(__name__)

# ----------------------------------------------------------


def onWorkerReady(name):
    logger.info("Worker '%s' is ready", name)
    try:
        worker = Worker.objects.get(name=name)
    except Worker.DoesNotExist:
        logger.error("No worker named '%s' found in the database", name)
    else:
        with transaction.atomic():
            worker.active = True
            worker.info = "Connected to the scheduler"
            worker.save()


# ----------------------------------------------------------


def onWorkerGone(name):
    logger.info("Worker '%s' is gone", name)
    try:
        worker = Worker.objects.get(name=name)
    except Worker.DoesNotExist:
        logger.error("No worker named '%s' found in the database", name)
    else:
        with transaction.atomic():
            worker.active = False
            worker.info = "Disconnected from the scheduler"
            worker.save()


class Command(BaseCommand):

    help = "Start zmq broker"

    def add_arguments(self, parser):
        parser.add_argument(
            "--port",
            "-p",
            type=int,
            dest="port",
            default=5555,
            help="Port of the broker",
        )

    def handle(self, *args, **options):
        verbosity = int(options["verbosity"])
        logger = setup_logging(verbosity, "Broker", __name__)

        if verbosity >= 1:
            if verbosity == 1:
                logger.setLevel(logging.INFO)
            elif verbosity >= 2:
                logger.setLevel(logging.DEBUG)

        logger.info("starting broker")

        address = "tcp://*:{}".format(options["port"])
        broker = BeatComputationBroker(verbosity >= 2)
        broker.set_worker_callbacks(onWorkerReady, onWorkerGone)
        broker.bind(address)
        broker.mediate()
        broker.purge_workers()

        logger.info("broker stopped")
