#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

###############################################################################
#                                                                             #
# Copyright (c) 2019 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.web module of the BEAT platform.              #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################

import os
import multiprocessing
import tempfile
import time

from unittest import skipIf

from django.test import TransactionTestCase
from django.conf import settings
from django.core.management import call_command
from django import db

from beat.core.utils import find_free_port

from beat.web.backend.models.worker import Worker
from beat.web.backend.tests.common import BackendUtilitiesMixin
from beat.web.backend.tests.common import ONE_QUEUE_TWO_WORKERS
from beat.web.backend.utils import setup_backend
from beat.web.backend.helpers import schedule_experiment
from beat.web.backend.helpers import cancel_experiment
from beat.web.experiments.models import Experiment
from beat.web.experiments.models import Block


def start_broker(port, verbosity=0):
    db.connections.close_all()
    call_command("broker", port=port, verbosity=verbosity)


def start_worker(worker_name, broker_address, prefix, cache, verbosity=0):
    call_command(
        "worker",
        name=worker_name,
        broker_address=broker_address,
        prefix=prefix,
        cache=cache,
        verbosity=verbosity,
    )


def start_scheduler(broker_address, verbosity=0):
    db.connections.close_all()
    call_command("scheduler", broker_address=broker_address, verbosity=verbosity)


@skipIf(settings.RUNNING_ON_CI, "Not runnable on the CI")
class WorkerRegistration(TransactionTestCase):
    def setUp(self):
        db_worker = Worker(name="test_worker1", active=False, cores=1, memory=12)
        db_worker.save()
        db_worker2 = Worker(name="test_worker2", active=False, cores=1, memory=12)
        db_worker2.save()

        self.db_workers = [db_worker, db_worker2]
        self.port = find_free_port()
        self.broker_address = "tcp://localhost:{}".format(self.port)

    def test_worker_registration(self):
        prefix = tempfile.TemporaryDirectory(prefix="beat.scheduling")
        cache = os.path.join(prefix.name, "cache")
        os.mkdir(cache)

        broker = multiprocessing.Process(target=start_broker, args=(self.port, 0))
        workers = []
        for db_worker in self.db_workers:
            worker = multiprocessing.Process(
                target=start_worker,
                args=(db_worker.name, self.broker_address, prefix.name, cache, 0),
            )
            workers.append(worker)

        broker.start()

        for worker in workers:
            worker.start()

        max_rounds = 5
        for db_worker in self.db_workers:
            db_worker.refresh_from_db()

        while max_rounds > 0 and not all([worker.active for worker in self.db_workers]):
            max_rounds -= 1
            worker.join(5)  # give time to have heartbeat sent and gone detected
            for db_worker in self.db_workers:
                db_worker.refresh_from_db()

        was_active = all([db_worker.active for worker in self.db_workers])

        for worker in workers:
            worker.terminate()
            worker.join()

        max_rounds = 5
        while max_rounds > 0 and not all(
            [not worker.active for worker in self.db_workers]
        ):
            max_rounds -= 1
            broker.join(5)  # give time to have heartbeat sent and gone detected
            for db_worker in self.db_workers:
                db_worker.refresh_from_db()

        broker.terminate()
        broker.join()

        self.assertTrue(was_active)
        self.assertTrue(all([not db_worker.active for worker in self.db_workers]))


# ----------------------------------------------------------


@skipIf(settings.RUNNING_ON_CI, "Not runnable on the CI")
class TestSchedulerBase(TransactionTestCase, BackendUtilitiesMixin):
    def __init__(self, methodName="runTest"):
        super(TestSchedulerBase, self).__init__(methodName)
        self.port = find_free_port()
        self.broker_address = "tcp://localhost:{}".format(self.port)

        self.scheduler = None
        self.broker = None
        self.workers = {}

        self.command_verbosity = 3
        self.timeout = settings.DB_REFRESH_TIMEOUT  # Failure timeout

    @classmethod
    def setUpTestData(cls):
        cls.setup_test_data()

    def setUp(self):
        self.shutdown_everything()
        self.clean_cache()

        TestSchedulerBase.setup_test_data()
        setup_backend(ONE_QUEUE_TWO_WORKERS)
        Worker.objects.update(active=False)

    def tearDown(self):
        self.shutdown_everything()
        self.clean_cache()

    def start_scheduling(self, verbosity=None):
        if verbosity is None:
            verbosity = self.command_verbosity

        self.start_broker(verbosity)
        self.start_scheduler(verbosity)

    def shutdown_everything(self):
        for name in list(self.workers.keys()):
            self.stop_worker(name)

        self.workers = {}

        self.stop_scheduler()
        self.stop_broker()

    def start_scheduler(self, verbosity=None):
        if verbosity is None:
            verbosity = self.command_verbosity

        self.scheduler = multiprocessing.Process(
            target=start_scheduler, args=(self.broker_address, verbosity)
        )
        self.scheduler.start()

    def stop_scheduler(self):
        if self.scheduler is not None:
            self.scheduler.terminate()
            self.scheduler.join()
            self.scheduler = None

    def start_broker(self, verbosity=None):
        if verbosity is None:
            verbosity = self.command_verbosity

        self.broker = multiprocessing.Process(
            target=start_broker, args=(self.port, verbosity)
        )
        self.broker.start()

    def stop_broker(self):
        if self.broker is not None:
            self.broker.terminate()
            self.broker.join()
            self.broker = None

    def start_worker(self, name, verbosity=None):
        if verbosity is None:
            verbosity = self.command_verbosity

        worker = multiprocessing.Process(
            target=start_worker,
            args=(
                name,
                self.broker_address,
                settings.PREFIX,
                settings.CACHE_ROOT,
                verbosity,
            ),
        )

        worker.start()

        self.workers[name] = worker

    def stop_worker(self, name):
        worker = self.workers.pop(name, None)

        if worker is not None:
            worker.terminate()
            worker.join()
            del worker

    def check_worker_status(self, name, active, timeout=10):
        start = time.time()
        worker = Worker.objects.get(name=name)

        while (
            worker.active != active
        ):  # Worker.objects.filter(name=name, active=active).count() == 0:
            if time.time() - start > timeout:
                self.shutdown_everything()
                self.assertTrue(False)
            worker.refresh_from_db()


# ----------------------------------------------------------


@skipIf(settings.RUNNING_ON_CI, "Not runnable on the CI")
class TestConnection(TestSchedulerBase):
    def test_worker_connection(self):
        self.start_scheduling()

        self.assertEqual(Worker.objects.filter(active=True).count(), 0)

        self.start_worker("node1")

        self.check_worker_status("node1", True)

        self.assertEqual(Worker.objects.filter(active=True).count(), 1)

    def test_worker_disconnection(self):
        self.start_scheduling()
        self.start_worker("node1")

        self.check_worker_status("node1", True)

        self.workers["node1"].terminate()
        self.workers["node1"].join()
        del self.workers["node1"]

        self.check_worker_status("node1", False)

    def test_two_workers_connection(self):
        self.start_scheduling()

        self.assertEqual(Worker.objects.filter(active=True).count(), 0)

        self.start_worker("node1")
        self.start_worker("node2")

        self.check_worker_status("node1", True)
        self.check_worker_status("node2", True)

        self.assertEqual(Worker.objects.filter(active=True).count(), 2)

    def test_scheduler_last(self):
        self.start_worker("node1")

        time.sleep(1)

        self.start_scheduling()

        self.check_worker_status("node1", True)

    def test_broker_shutdown(self):
        self.start_broker()

        self.assertEqual(Worker.objects.filter(active=True).count(), 0)

        self.start_worker("node1")

        self.check_worker_status("node1", True)

        self.assertEqual(Worker.objects.filter(active=True).count(), 1)

        self.stop_broker()

        self.assertEqual(Worker.objects.filter(active=False).count(), 1)

        time.sleep(1)

        self.start_broker()

        self.check_worker_status("node1", True)

        self.assertEqual(Worker.objects.filter(active=True).count(), 1)


# ----------------------------------------------------------


@skipIf(settings.RUNNING_ON_CI, "Not runnable on the CI")
class TestExecution(TestSchedulerBase):
    def setUp(self):
        super(TestExecution, self).setUp()

        self.start_scheduling()
        self.start_worker("node1")
        self.start_worker("node2")
        self.check_worker_status("node1", True)
        self.check_worker_status("node2", True)

    def test_successful_experiment(self):
        fullname = "user/user/double/1/double"

        xp = Experiment.objects.get(name=fullname.split("/")[-1])

        self.prepare_databases(xp.declaration)

        schedule_experiment(xp)
        xp.refresh_from_db()

        start = time.time()
        while xp.status != Experiment.DONE:
            self.assertTrue(time.time() - start < self.timeout)
            xp.refresh_from_db()

        b1 = xp.blocks.get(name="echo1")
        b2 = xp.blocks.get(name="echo2")
        b3 = xp.blocks.get(name="analysis")

        self.assertEqual(b1.status, Block.DONE)
        self.assertEqual(b2.status, Block.DONE)
        self.assertEqual(b3.status, Block.DONE)

    def test_successful_splitted_experiment(self):
        fullname = "user/user/double/1/double_split_2"

        xp = Experiment.objects.get(name=fullname.split("/")[-1])

        self.prepare_databases(xp.declaration)

        schedule_experiment(xp)
        xp.refresh_from_db()

        start = time.time()
        while xp.status != Experiment.DONE:
            self.assertTrue(time.time() - start < self.timeout)
            xp.refresh_from_db()

        b1 = xp.blocks.get(name="echo1")
        b2 = xp.blocks.get(name="echo2")
        b3 = xp.blocks.get(name="analysis")

        self.assertEqual(b1.status, Block.DONE)
        self.assertEqual(b2.status, Block.DONE)
        self.assertEqual(b3.status, Block.DONE)

    def test_two_similar_experiments(self):
        fullname1 = "user/user/double/1/double"
        fullname2 = "user/user/double/1/double_split_2"

        xp1 = Experiment.objects.get(name=fullname1.split("/")[-1])
        xp2 = Experiment.objects.get(name=fullname2.split("/")[-1])

        self.prepare_databases(xp1.declaration)
        self.prepare_databases(xp2.declaration)

        schedule_experiment(xp1)
        schedule_experiment(xp2)

        xp1.refresh_from_db()
        xp2.refresh_from_db()

        start = time.time()
        while (xp1.status != Experiment.DONE) or (xp2.status != Experiment.DONE):
            self.assertTrue(time.time() - start < self.timeout)
            xp1.refresh_from_db()
            xp2.refresh_from_db()

        self.assertEqual(xp1.start_date, xp2.start_date)
        self.assertEqual(xp1.end_date, xp2.end_date)

        b1 = xp1.blocks.get(name="echo1")
        b2 = xp2.blocks.get(name="echo1")

        self.assertEqual(b1.status, Block.DONE)
        self.assertEqual(b2.status, Block.DONE)
        self.assertEqual(b1.start_date, b2.start_date)
        self.assertEqual(b1.end_date, b2.end_date)

        b1 = xp1.blocks.get(name="echo2")
        b2 = xp2.blocks.get(name="echo2")

        self.assertEqual(b1.status, Block.DONE)
        self.assertEqual(b2.status, Block.DONE)
        self.assertEqual(b1.start_date, b2.start_date)
        self.assertEqual(b1.end_date, b2.end_date)

        b1 = xp1.blocks.get(name="analysis")
        b2 = xp2.blocks.get(name="analysis")

        self.assertEqual(b1.status, Block.DONE)
        self.assertEqual(b2.status, Block.DONE)
        self.assertEqual(b1.start_date, b2.start_date)
        self.assertEqual(b1.end_date, b2.end_date)

    def test_two_different_experiments(self):
        fullname1 = "user/user/single/1/single"
        fullname2 = "user/user/single/1/single_add"

        xp1 = Experiment.objects.get(name=fullname1.split("/")[-1])
        xp2 = Experiment.objects.get(name=fullname2.split("/")[-1])

        self.prepare_databases(xp1.declaration)
        self.prepare_databases(xp2.declaration)

        schedule_experiment(xp1)
        schedule_experiment(xp2)

        xp1.refresh_from_db()
        xp2.refresh_from_db()

        start = time.time()
        while (xp1.status != Experiment.DONE) or (xp2.status != Experiment.DONE):
            self.assertTrue(time.time() - start < self.timeout)
            xp1.refresh_from_db()
            xp2.refresh_from_db()

        self.assertNotEqual(xp1.start_date, xp2.start_date)
        self.assertNotEqual(xp1.end_date, xp2.end_date)

    def test_two_delayed_similar_experiments(self):
        fullname1 = "user/user/double/1/double"
        fullname2 = "user/user/double/1/double_split_2"

        xp1 = Experiment.objects.get(name=fullname1.split("/")[-1])
        xp2 = Experiment.objects.get(name=fullname2.split("/")[-1])

        self.prepare_databases(xp1.declaration)
        self.prepare_databases(xp2.declaration)

        schedule_experiment(xp1)

        start = time.time()
        while xp1.status != Experiment.RUNNING:
            self.assertTrue(time.time() - start < self.timeout)
            xp1.refresh_from_db()

        schedule_experiment(xp2)
        # xp2.refresh_from_db()

        start = time.time()
        while xp1.status != Experiment.DONE or xp2.status != Experiment.DONE:
            self.assertTrue(time.time() - start < self.timeout)
            xp1.refresh_from_db()
            xp2.refresh_from_db()

    def test_failed_experiment(self):
        fullname = "user/user/single/1/single_error"

        xp = Experiment.objects.get(name=fullname.split("/")[-1])

        self.prepare_databases(xp.declaration)

        schedule_experiment(xp)
        xp.refresh_from_db()

        start = time.time()
        while xp.status != Experiment.FAILED:
            self.assertTrue(time.time() - start < self.timeout)
            xp.refresh_from_db()

        b1 = xp.blocks.get(name="echo")
        b2 = xp.blocks.get(name="analysis")

        self.assertEqual(b1.status, Block.FAILED)
        self.assertEqual(b2.status, Block.CANCELLED)

    def test_failed_splitted_experiment(self):
        fullname = "user/user/double/1/double_error_split_2"

        xp = Experiment.objects.get(name=fullname.split("/")[-1])

        self.prepare_databases(xp.declaration)

        schedule_experiment(xp)
        xp.refresh_from_db()

        start = time.time()
        while xp.status != Experiment.FAILED:
            self.assertTrue(time.time() - start < self.timeout)
            xp.refresh_from_db()

        b1 = xp.blocks.get(name="echo1")
        b2 = xp.blocks.get(name="echo2")
        b3 = xp.blocks.get(name="analysis")

        self.assertEqual(b1.status, Block.DONE)
        self.assertEqual(b2.status, Block.FAILED)
        self.assertEqual(b3.status, Block.CANCELLED)

    def test_failed_mirror_experiments(self):
        fullname1 = "user/user/double/1/double_error"
        fullname2 = "user/user/double/1/double_error_split_2"

        xp1 = Experiment.objects.get(name=fullname1.split("/")[-1])
        xp2 = Experiment.objects.get(name=fullname2.split("/")[-1])

        self.prepare_databases(xp1.declaration)
        self.prepare_databases(xp2.declaration)

        schedule_experiment(xp1)
        schedule_experiment(xp2)

        xp1.refresh_from_db()
        xp2.refresh_from_db()

        start = time.time()
        while xp1.status != Experiment.FAILED or xp2.status != Experiment.FAILED:
            self.assertTrue(time.time() - start < self.timeout)
            xp1.refresh_from_db()
            xp2.refresh_from_db()

        b1 = xp1.blocks.get(name="echo1")
        b2 = xp1.blocks.get(name="echo2")
        b3 = xp1.blocks.get(name="analysis")

        self.assertEqual(b1.status, Block.DONE)
        self.assertEqual(b2.status, Block.FAILED)
        self.assertEqual(b3.status, Block.CANCELLED)

        b1 = xp2.blocks.get(name="echo1")
        b2 = xp2.blocks.get(name="echo2")
        b3 = xp2.blocks.get(name="analysis")

        self.assertEqual(b1.status, Block.DONE)
        self.assertEqual(b2.status, Block.FAILED)
        self.assertEqual(b3.status, Block.CANCELLED)


# ----------------------------------------------------------


@skipIf(settings.RUNNING_ON_CI, "Not runnable on the CI")
class TestCancellation(TestSchedulerBase):
    def setUp(self):
        super(TestCancellation, self).setUp()

        self.start_scheduling(0)
        self.start_worker("node1", 0)
        self.check_worker_status("node1", True)
        self.start_worker("node2", 0)
        self.check_worker_status("node2", True)

    def process(self, experiment_name, block_name=None):
        xp_name = experiment_name.split("/")[-1]

        xp = Experiment.objects.get(name=xp_name)

        self.prepare_databases(xp.declaration)

        schedule_experiment(xp)
        xp.refresh_from_db()

        start = time.time()
        while xp.status != Experiment.RUNNING:
            self.assertTrue(time.time() - start < self.timeout)
            xp.refresh_from_db()

        if block_name is not None:
            block = xp.blocks.get(name=block_name)

            start = time.time()
            while block.status != Block.PROCESSING:
                self.assertTrue(time.time() - start < self.timeout)
                block.refresh_from_db()

        cancel_experiment(xp)
        xp.refresh_from_db()

        start = time.time()
        while xp.status != Experiment.PENDING:
            self.assertTrue(time.time() - start < self.timeout)
            xp.refresh_from_db()

        self.assertTrue(xp.blocks.filter(status=Block.CANCELLED).count() > 0)
        self.assertEqual(xp.blocks.filter(status=Block.PENDING).count(), 0)
        self.assertEqual(xp.blocks.filter(status=Block.PROCESSING).count(), 0)

    def process2(self, experiment_name1, experiment_name2, cancel_index=0):
        xp1 = Experiment.objects.get(name=experiment_name1.split("/")[-1])
        xp2 = Experiment.objects.get(name=experiment_name2.split("/")[-1])

        self.prepare_databases(xp1.declaration)
        self.prepare_databases(xp2.declaration)

        schedule_experiment(xp1)
        schedule_experiment(xp2)
        xp1.refresh_from_db()
        xp2.refresh_from_db()

        start = time.time()
        while xp1.status != Experiment.RUNNING or xp2.status != Experiment.RUNNING:
            self.assertTrue(time.time() - start < self.timeout)
            xp1.refresh_from_db()
            xp2.refresh_from_db()

        if cancel_index == 0:
            xp_to_cancel = xp1
            xp_to_finish = xp2
        else:
            xp_to_cancel = xp2
            xp_to_finish = xp1

        cancel_experiment(xp_to_cancel)

        xp_to_cancel.refresh_from_db()
        xp_to_finish.refresh_from_db()

        start = time.time()
        while xp_to_cancel.status != Experiment.PENDING:
            self.assertTrue(time.time() - start < self.timeout)
            xp_to_cancel.refresh_from_db()

        xp_to_finish.refresh_from_db()

        start = time.time()
        while xp_to_finish.status != Experiment.DONE:
            self.assertTrue(time.time() - start < self.timeout)
            xp_to_finish.refresh_from_db()

        self.assertTrue(xp_to_cancel.blocks.filter(status=Block.CANCELLED).count() > 0)
        self.assertEqual(xp_to_cancel.blocks.filter(status=Block.PENDING).count(), 0)
        self.assertEqual(xp_to_cancel.blocks.filter(status=Block.PROCESSING).count(), 0)

        self.assertEqual(xp_to_finish.blocks.filter(status=Block.CANCELLED).count(), 0)
        self.assertEqual(xp_to_finish.blocks.filter(status=Block.PENDING).count(), 0)
        self.assertEqual(xp_to_finish.blocks.filter(status=Block.PROCESSING).count(), 0)

    def test_one_split_running(self):
        self.process("user/user/single/1/single_sleep_4")

    def test_one_split_of_two_in_a_block_running(self):
        self.stop_worker("node2")
        self.check_worker_status("node2", False)
        self.process("user/user/double/1/double_sleep_split_2", block_name="echo2")

    def test_two_splits_of_same_block_running(self):
        self.process("user/user/double/1/double_sleep_split_2", block_name="echo2")

    def test_two_splits_of_different_blocks_running(self):
        self.process("user/user/triangle/1/triangle_sleep_4")

    def test_mirror_block(self):
        self.stop_worker("node2")
        self.check_worker_status("node2", False)
        self.process2(
            "user/user/single/1/single_sleep_4",
            "user/user/single/1/single_sleep_5",
            cancel_index=1,
        )

    def test_mirrored_block(self):
        self.stop_worker("node2")
        self.check_worker_status("node2", False)
        self.process2(
            "user/user/single/1/single_sleep_4",
            "user/user/single/1/single_sleep_5",
            cancel_index=0,
        )
